/*
 * Copyright 2008  Rice University
 * All Rights Reserved
 * 
 *       Created by Erion Plaku, Lydia Kavraki
 *       Physical and Biological Computing Group
 *       Department of Computer Science
 *       Rice University
 */

/**
 *@file OOPSMPPRM.H
 *@author Erion Plaku 
 *@date First version: March 15, 2007.  Last update: March 15, 2007.
 *@brief Implementation of the Probabilistic RoadMap (<em>PRM</em>)
 *       method  
 *
 *@remark
 * Algorithm described in Kavraki, L. E., Svestka, P., Latombe, J.-C.,
 * and Overmars, M.. Probabilistic Roadmaps for Path Planning in High
 * Dimensional Configuration Spaces. IEEE Transactions on Robotics and
 * Automation, 12(4):566-580, 1996.  
 */

#ifndef OOPSMP_MOTION_PLANNERS_PRM__OOPSMP_PRM_H_
#define OOPSMP_MOTION_PLANNERS_PRM__OOPSMP_PRM_H_

#include "MotionPlanners/OOPSMPMotionPlanner.H"
#include "Utils/DataStructures/Proximity/OOPSMPProximity.H"
#include "Utils/DataStructures/Graph/OOPSMPGraph.H"
#include "Utils/DataStructures/SearchGraph/OOPSMPSearchGraph.H"
#include "Utils/DataStructures/DisjointSet/OOPSMPDisjointSet.H"

namespace OOPSMP
{
    /// @cond
    /**
     *@author Erion Plaku
     *@brief Macro for class, class pointer, and const class pointer
     *       declarations 
     */  
    ForwardClassDeclaration(PRM);
    /// @endcond
    
    /**
     *@author Erion Plaku
     *@brief Implementation of the Probabilistic RoadMap
     *       (<em>PRM</em>) method  
     *
     *@remark
     * Algorithm described in Kavraki, L. E., Svestka, P., Latombe, J.-C.,
     * and Overmars, M.. Probabilistic Roadmaps for Path Planning in High
     * Dimensional Configuration Spaces. IEEE Transactions on Robotics and
     * Automation, 12(4):566-580, 1996.  
     *
     *@par Plug-and-play functionality:
     *  @code
     *  BeginImplementInstanceFactory(PRM, MotionPlannerFactory);
     *    RegisterFnFactory(allowCycles, FACTORY_VAL_ARG(const bool, 0));
     *    RegisterFnFactory(setNrNeighborsSample,    FACTORY_VAL_ARG(int, 0));
     *    RegisterFnFactory(setNrNeighborsQuery,     FACTORY_VAL_ARG(int, 0));
     *    RegisterFnFactory(setFractionSamplingTime, FACTORY_VAL_ARG(double, 0));
     *    RegisterFnFactory(setMinMaxNrSamples,      FACTORY_VAL_ARG(int, 0), FACTORY_VAL_ARG(int, 1));
     *    RegisterFnFactory(setProximity,            FACTORY_PTR_ARG(ProximityKeyPointer_t, 0));
     *  EndImplementInstanceFactory(PRM);
     *  @endcode 
     */  
    class PRM : public MotionPlanner
    {
    public:
	/**
	 *@author Erion Plaku
	 *@brief Initialize data elements to appropriate default
	 *       values 
         */
	PRM(void);
	
	/**
	 *@author Erion Plaku
	 *@brief Free the memory allocated to different data elements 
         */	
	virtual ~PRM(void);
	
	/**
	 *@name Data associated with PRM vertices
	 *@{
	 */

	/// @cond
	/**
	 *@author Erion Plaku
	 *@brief Macro for class, class pointer, and const class
	 *       pointer declarations 
	 */  
	ForwardClassDeclaration(VertexData);
	/// @endcond
		
	/**
	 *@author Erion Plaku
	 *@brief Vertices of PRM are states
	 */  
	class VertexData : public GraphVertexData
	{
	public:
	    friend class PRM;
	    friend class VisPRM;

	    /**
	     *@author Erion Plaku
	     *@brief Initialize data elements to appropriate default
	     *       values 
	     */	    
	    VertexData(const int id = -1, State_t x = NULL) : GraphVertexData(), m_id(id), m_x(x)
	    {
	    }

	    /**
	     *@author Erion Plaku
	     *@brief Free the memory allocated to different data
	     *       elements 
	     */	
	    virtual ~VertexData(void)
	    {		
		if(m_x)
		    free(m_x);
	    }

	    /**
	     *@name Access vertex data
	     *@{
	     */

	    /**
	     *@author Erion Plaku
	     *@brief Get id associated with the vertex
	     */		    
	    int getId(void) const
	    {
		return m_id;
	    }

	    /**
	     *@author Erion Plaku
	     *@brief Get state associated with the vertex
	     */		    
	    State_t getState(void) const
	    {
		return m_x;
	    }

	    /**
	     *END Access vertex data
	     *@}
	     */

	    /**
	     *@name Setup vertex data
	     *@{
	     */
 
	    /**
	     *@author Erion Plaku
	     *@brief Set id associated with the vertex
	     *@param id vertex id
	     */		    
	    virtual void setId(const int id)
	    {
		m_id = id;
	    }
	    
	    /**
	     *@author Erion Plaku
	     *@brief Set state associated with the vertex
	     *@param x state
	     *
	     *@remark
	     *  The function performs only pointer assignment.  Note
	     *  that the memory associated with <em>x</em> is freed
	     *  when the destructor is called. 
	     */		    
	    virtual void setState(State_t x)
	    {
		m_x = x;
	    }

	    /**
	     *END Setup vertex data
	     *@}
	     */

	protected:	    
	    /**
	     *@author Erion Plaku
	     *@brief Id associated with the vertex
	     */		    
	    int  m_id;

	    /**
	     *@author Erion Plaku
	     *@brief State associated with the vertex
	     */		    
	    State_t m_x;
	};

	/**
	 *END Data associated with PRM vertices
	 *@}
	 */

	/**
	 *@name Data associated with PRM edges
	 *@{
	 */

	/// @cond
	/**
	 *@author Erion Plaku
	 *@brief Macro for class, class pointer, and const class
	 *       pointer declarations 
	 */  
	ForwardClassDeclaration(EdgeData);
	/// @endcond

	/**
	 *@author Erion Plaku
	 *@brief A PRM edge consists of a path that connects the first
	 *       vertex to the second vertex 
	 */  
	class EdgeData : public GraphEdgeData
	{
	public:
	    friend class PRM;

	    /**
	     *@author Erion Plaku
	     *@brief Initialize data elements to appropriate default
	     *       values 
	     */	    
	    EdgeData(Path_t path = NULL) : GraphEdgeData(), m_path(path)
	    {
	    }

	    /**
	     *@author Erion Plaku
	     *@brief Free the memory allocated to different data
	     *       elements 
	     */	
    	    virtual ~EdgeData(void)
	    {
		if(m_path)
		    delete m_path;
	    }

	    /**
	     *@name Access edge data
	     *@{
	     */

	    /**
	     *@author Erion Plaku
	     *@brief Get path associated with the edge
	     */		    
	    Path_t getPath(void) const
	    {
		return m_path;
	    }

	    /**
	     *END Access edge data
	     *@}
	     */

	    /**
	     *@name Setup edge data
	     *@{
	     */

	    /**
	     *@author Erion Plaku
	     *@brief Set path associated with the edge
	     *@param path path
	     *
	     *@remark
	     *  The function performs only pointer assignment.  Note
	     *  that the memory associated with <em>path</em> is freed
	     *  when the destructor is called. 
	     */		    
	    virtual void setPath(Path_t path)
	    {
		m_path = path;
	    }

	    /**
	     *END Setup edge data
	     *@}
	     */
	    
	protected:
	    /**
	     *@author Erion Plaku
	     *@brief Path associated with the edge
	     */		    	    
	    Path_t m_path;
	};

	/**
	 *END Data associated with PRM edges
	 *@}
	 */

	/**
	 *@name Access motion planner data
	 *@{
	 */
	
	/**
	 *@author Erion Plaku
	 *@brief Indicate whether the current motion planning query
	 *       has been solved or not 
         */
	bool isProblemSolved(void) const
	{
	    return m_isSolved;
	}
	
	/**
	 *@author Erion Plaku
	 *@brief Get fraction of preprocessing time allocated to state
	 *       sampling 
         */
	double getFractionSamplingTime(void) const
	{
	    return m_fractionSamplingTime;
	}

	/**
	 *@author Erion Plaku
	 *@brief Get minimum number of samples that should be generated --
	 *       a negative value indicates no lower bound
         */
	int getMinNrSamples(void) const
	{
	    return m_minNrSamples;
	}

	/**
	 *@author Erion Plaku
	 *@brief Get maximum number of samples that should be generated --
	 *       a negative value indicates no upper bound
         */
	int getMaxNrSamples(void) const
	{
	    return m_maxNrSamples;
	}
	
	/**
	 *@author Erion Plaku
	 *@brief Get roadmap data structure
         */
	Graph_t getRoadmap(void) const
	{
	    return m_roadmap;
	}
	
	/**
	 *@author Erion Plaku
	 *@brief Indicate whether cycles are allowed in the roadmap or
	 *       not 
	 *@returns
	 *  <em>true</em> iff cycles are allowed in the roadmap
         */
	bool areCyclesAllowed(void) const
	{
	    return m_allowCycles;
	}
		
	/**
	 *@author Erion Plaku
	 *@brief Get data structure for computing neighbors 
	 */
	ProximityKeyPointer_t getProximity(void) const
	{
	    return m_proximity;
	}
	
	/**
	 *@author Erion Plaku
	 *@brief Get number of neighbors to try to connect each state
	 *       in the roadmap  
         */
	int getNrNeighborsSample(void) const
	{
	    return m_nrNeighborsSample;
	}
	
	/**
	 *@author Erion Plaku
	 *@brief Get number of neighbors to try to connect each query
	 *       state to the states in the roadmap  
         */
	int getNrNeighborsQuery(void) const
	{
	    return m_nrNeighborsQuery;
	}

	/**
	 *@author Erion Plaku
	 *@brief Get the search-graph data structure that is used for
	 *       searching the roadmap graph 
	 *
	 *@remarks
	 *  The memory associated with <em>sg</em> should not be freed 
	 *  by the calling function. This memory will be freed by the 
	 *  destructor of PRM. 
         */
	SearchGraph_t getSearchGraph(void) const
	{
	    return m_searchGraph;
	}
		
	/**
	 *END Access motion planner data
	 *@}
	 */

        /**
	 *@name Setup motion planner
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Set fraction of preprocessing time allocated to state
	 *       sampling 
	 *@param ft fraction of preprocessing time allocated to state
	 *       sampling  
         */
	virtual void setFractionSamplingTime(const double ft)
	{
	    m_fractionSamplingTime = ft;
	}

	/**
	 *@author Erion Plaku
	 *@brief Set min and maximum number of samples that should be generated --
	 *       negative values indicates no lower/upper bounds
	 *@param min minimum number of samples
	 *@param max maximum number of samples
         */
	virtual void setMinMaxNrSamples(const int min, const int max)
	{
	    m_minNrSamples = min;
	    m_maxNrSamples = max;
	}
	
	/**
	 *@author Erion Plaku
	 *@brief Indicate whether or not roadmap cycles should be allowed 
	 *@param allow boolean flag
         */
	virtual void allowCycles(const bool allow)
	{
	    m_allowCycles = allow;
	}

	/**
	 *@author Erion Plaku
	 *@brief Set the proximity data structure that should be used
	 *       for neighbor computations 
	 *@param proximity pointer to a proximity data structure  
	 *
	 *@remarks
	 *  The memory associated with <em>proximity</em> should not
	 *  be freed by the calling function. This memory will be
	 *  freed by the destructor of PRM. 
         */
	virtual void setProximity(ProximityKeyPointer_t proximity)
	{
	    m_proximity = proximity;
	}

	/**
	 *@author Erion Plaku
	 *@brief Set number of neighbors to try to connect each state
	 *       in the roadmap  
	 *@param k number of neighbors
         */
	virtual void setNrNeighborsSample(const int k)
	{
	    m_nrNeighborsSample = k;
	}

	/**
	 *@author Erion Plaku
	 *@brief Set number of neighbors to try to connect each query
	 *       state to the states in the roadmap  
	 *@param k number of neighbors
         */
	virtual void setNrNeighborsQuery(const int k)
	{
	    m_nrNeighborsQuery = k;
	}

	/**
	 *@author Erion Plaku
	 *@brief Set the search-graph data structure that should be
	 *       used for searching the roadmap graph 
	 *@param sg pointer to a data structure for searching the
	 *       roadmap graph 
	 *
	 *@remarks
	 *  The memory associated with <em>sg</em> should not be freed
	 *  by the calling function. This memory will be freed by the
	 *  destructor of PRM. 
         */
	virtual void setSearchGraph(SearchGraph_t sg)
	{
	    if(m_searchGraph != sg && m_searchGraph)
		delete m_searchGraph;
	    m_searchGraph = sg;
	}
		
        /**
	 *END Setup motion planner
	 *@}
	 */

        /**
	 *@name Use motion planner to solve one or more queries
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Solve multiple motion planning queries
	 *@param ptime maximum amount of preprocessing time allocated
	 *       to motion planner 
	 *@param stime maximum amount of solution time allocated to
	 *       motion planner 
	 *@param queries pointer to an instance of Queries, which
	 *       contains multiple queries 
	 *@param perm a random permutation of numbers <em>0</em> to
	 *       <em>n - 1</em>, where <em>n</em> is the number of
	 *       queries 
	 *@par Description
	 *   Motion planner attempts to solve the different multiple
	 *   queries specified by <em>queries</em> using at most
	 *   <em>ptime</em> and <em>stime</em> seconds for
	 *   preprocessing and actual query solving, respectively. The
	 *   queries are solved in the random order specified by
	 *   <em>perm</em>. Solution paths, if found, are stored with
	 *   each query. \n\n
	 *   PRM uses the preprocessing time to construct the roadmap
	 *   and the solution time to connect queries to the roadmap.  
	 *   
	 *@remark
	 *  - If <em>perm</em> is <em>NULL</em>, the identity order is
	 *    used. 
	 *  - The function performs only pointer assignment of
	 *    queries.  Therefore the memory associated with <em>
	 *    queries </em> should not be freed while the motion
	 *    planner is in use. Note that the motion planner does not
	 *    free the memory associated with <em> queries</em>, so it
	 *    is the responsibility of the user program to free this
	 *    memory at the end. 
	 *  .
         */
	void solveMultipleQueries(const double ptime, 
				  const double stime, 
				  Queries_t    queries, 
				  const int *perm = NULL)
	{
	    solveMultipleQueries_PreprocessOnce(ptime, stime, queries, perm);
	}

        /**
	 *END Use motion planner to solve one or more queries
	 *@}
	 */
	
        /**
	 *@name Solve the current motion planner query
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Clear the data structures and other data elements in
	 *       preparation for solving motion planning queries 
	 *@par Description
	 * Clear the roadmap data structure, deleting the graph and
	 * the vertex data and edge data associated with the vertices
	 *  and edges of the graph, respectively. 
         */
	void clear(void);


	/**
	 *@author Erion Plaku
	 *@brief Initialize or construct data structures that will
	 *       facilitate the solution of motion planning queries 
	 *@param ptime maximum amount of preprocessing time allocated
	 *       to motion planner 
 	 *@par Description
	 *   Motion planner uses at most <em>ptime</em> seconds for
	 *   preprocessing. PRM uses the preprocessing time to
	 *   construct the roadmap.  
	 */
	void preprocess(const double ptime);

	/**
	 *@author Erion Plaku
	 *@brief Solve the current motion planning qury
	 *@param stime maximum amount of solution time allocated to
	 *       motion planner 
	 *
	 *@par Description
	 *   Motion planner attempts to solve the current motion
	 *   planning query using at most  <em>stime</em> seconds.  A
	 *   solution path, if found, is stored with the query. 
         */	
	bool solve(const double stime);
	
        /**
	 *END Solve the current motion planner query
	 *@}
	 */

	/**
	 *@name Compute solution path
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Compute the solution path for the current query -- if
	 *       the query is solved. 
	 *
	 *@remark
	 *  If the current query is solved, a solution path is
	 *  computed and stored as a data element of the current query
	 *  using the Query::setSolutionPath function. 
         */
	void computeSolutionPath(void);

	/**
	 *END Compute solution path
	 *@}
	 */

	/**
	 *@name Print motion planning data
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Print states and paths of the data structure
	 *       constructed to solve motion planning queries 
	 *@param dt time discretization value for printing paths 
	 *@param out pointer to a file open for writing
	 *
	 *@remark
	 *  States and paths are printed using the print functions in
	 *  CoreRobotData. Each path is printed as a sequence of
	 *  states obtained by discretizing the path at a <em>dt</em>
	 *  time interval.  
         */
	void printData(const double dt, FILE *out) const;

	/**
	 *END Print motion planning data
	 *@}
	 */
	
	/**
	 *@name State iterator
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Begin iteration of states generated by the motion
	 *       planning data structure 
 	 *@returns
	 *   Pointer to an instance of MotionPlanner::StateIterator
	 *   that can be used to iterate over all the states 
	 *   generated by the motion planning data structure. 
	 *   \n\n
	 *   A NULL pointer is returned if no states have been
	 *   generated. 
	 */
	MotionPlanner::StateIterator* beginStateIterator(void) const
	{
	    return (m_roadmap && m_roadmap->getNrVertices() > 0) ? new StateIterator() : NULL;
	}

	/**
	 *@author Erion Plaku
	 *@brief Access the state associated with the current value of
	 *       the iterator 
	 *@param si a non-NULL state iterator
	 *@remarks
	 *  The memory associated with the state returned by
	 *  getStateIterator should not be freed by the calling function. 
	 *  This memory -- if at all allocated -- will be freed by the
	 *  state iterator when the iteration reaches the end. 
	 */
	State_const_t getStateIterator(const MotionPlanner::StateIterator *si) const
	{
	    return dynamic_cast<VertexData_t>(m_roadmap->getVertexData(dynamic_cast<const StateIterator *>(si)->m_id))->m_x;
	}

	/**
	 *@author Erion Plaku
	 *@brief Advance the state iterator 
	 *@param si a non-NULL state iterator
	 *
	 *@returns
	 *  Advances the current iterator to the next state in the
	 *  iteration. NULL is returned if <em>si</em> points beyond
	 *  the last position in the state iteration and the memory
	 *  associated with <em>si</em> is freed. 
	 */	
	MotionPlanner::StateIterator* advanceStateIterator(MotionPlanner::StateIterator *si) const
	{
	    StateIterator *siter = dynamic_cast<StateIterator *>(si);
	    siter->m_id++;
	    if(siter->m_id >= m_roadmap->getNrVertices())
	    {
		delete siter;
		return NULL;
	    }
	    return siter;
	}

	/**
	 *END State iterator
	 *@}
	 */

	/**
	 *@name Path iterator
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Begin iteration of paths generated by the motion
	 *       planning data structure 
 	 *@returns
	 *   Pointer to an instance of MotionPlanner::PathIterator
	 *   that can be used to iterate over all the paths generated
	 *   by the motion planning data structure. \n\n
	 *   A NULL pointer is returned if no paths have been
	 *   generated. 
	 */
	MotionPlanner::PathIterator* beginPathIterator(void) const;

	/**
	 *@author Erion Plaku
	 *@brief Access the path associated with the current value of
	 *       the iterator 
	 *@param pi a non-NULL path iterator
	 *@remarks
	 *  The memory associated with the path returned by
	 *  getPathIterator should not be freed by the calling function. 
	 *  This memory -- if at all allocated -- will be freed by the
	 *  path iterator when the iteration reaches the end. 
	 */
	Path_const_t getPathIterator(const MotionPlanner::PathIterator *pi) const
	{
	    const PathIterator *piter = dynamic_cast<const PathIterator *>(pi);

	    return dynamic_cast<EdgeData_const_t>(piter->m_eout->getDataIterator(piter->m_eiter))->m_path;
	}

	/**
	 *@author Erion Plaku
	 *@brief Advance the path iterator 
	 *@param pi a non-NULL path iterator
	 *@returns
	 *  Advances the path iterator to the next path. NULL is
	 *  returned if <em>pi</em> points to the last position in the
	 *  path iteration and the memory associated with <em>si</em>
	 *  is freed. 
	 */
	MotionPlanner::PathIterator* advancePathIterator(MotionPlanner::PathIterator *pi) const;

	/**
	 *END Path iterator
	 *@}
	 */

	/**
	 *@name Generate and connect samples
	 *@{
	 */
	
        /**
	 *@author Erion Plaku
	 *@brief Generate and connect samples
	 *@param rtime maximum amount of time 
	 */
	virtual void generateAndConnectSamples(const double rtime);

	/**
	 *END Generate and connect samples
	 *@}
	 */

	/**
	 *@name Generate and add samples to roadmap
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Generate valid state samples
	 *@param rtime maximum amount of time used for generating
	 *       state samples 
	 *
	 *@par Description
	 *  Samples are generated using the
	 *  CoreRobotData::getValidStateSampler() state sampler. Each
	 *  valid state sample is added to the vertices of the
	 *  roadmap. This process is repeated until <em>rtime</em> is
	 *  exceeded. 
         */
	virtual void generateSamples(const double rtime);

        /**
	 *@author Erion Plaku
	 *@brief Generate one valid state sample
	 *@param rtime maximum amount of time used for generating
	 *       state sample 
	 *
	 *@returns 
	 * <em>true</em> iff a valid state sample is generated within
	 * <em>rtime</em> seconds 
	 *
	 *@par Description
	 *  The sample is generated using the
	 *  CoreRobotData::getValidStateSampler() state sampler. Each
	 *  valid state sample is added to the vertices of the
	 *  roadmap.   
	 */
	virtual bool generateOneSample(const double rtime);

        /**
	 *@author Erion Plaku
	 *@brief Generate <em>n</em> valid state samples
	 *@param n number of samples to generate
	 *@param rtime maximum time
	 */
	virtual void generateManySamples(const int n, const double rtime = HUGE_VAL);

	/**
	 *@author Erion Plaku
	 *@brief Add state as a roadmap vertex
	 *@param x state
         */
	virtual int addSample(State_t x);

	/**
	 *END Generate and add samples to roadmap
	 *@}
	 */

	/**
	 *@name Connect samples and add edges to roadmap
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Connect neighboring roadmap state samples with valid
	 *       paths 
	 *@param rtime maximum amount of time used for connecting
	 *       samples 
	 *
	 *@par Description
	 *  Attempts are made to connect each roadmap state sample to 
	 *  getNrNeighborsSample() neighbors, where neighbors are 
	 *  computed by the Proximity data structure and connections
	 *  are computed by using
	 *  LocalPlanner::generateValidForwardConnectPath. \n\n
	 *  This is an iterative method.  An index is maintained to
	 *  indicate the current index of the roadmap vertex which
	 *  will be connected to its neighbors. If after attempting
	 *  these connections,  there is time left, the current index
	 *  is incremented by one and this step is repeated. 
         */
	virtual void connectSamples(const double rtime);

	/**
	 *@author Erion Plaku
	 *@brief Connect current sample to several of its neigbors
	 *       with valid paths 
	 *@param idFrom index of roadmap vertex
	 *@param k      number of neighbors
	 *
	 *@par Description
	 *  Attempts are made to connect the state sample associated 
	 *  with the vertex with index <em>idFrom</em> to <em>k</em>
	 *  neighbors, where neighbors are computed by the Proximity
	 *  data structure and connections are computed by using
	 *  LocalPlanner::generateValidForwardConnectPath. 
         */
	virtual void connectSample(const int idFrom, const int k);

	/**
	 *@author Erion Plaku
	 *@brief Determine if the edge connecting the indicated
	 *       vertices should be added to the roadmap 
	 *@param idFrom index of first vertex of roadmap edge
	 *@param idTo index of second vertex of roadmap edge
	 *
	 *@par Description
	 *  The edge connecting vertex <em>idFrom</em> to vertex
	 *  <em>idTo</em> should be added to the roadmap if cycles are
	 *  allowed or if there is no path from vertex <em>idFrom</em>
	 *  to <em>idTo</em> 
         */
	virtual bool shouldAddEdge(const int idFrom, const int idTo);

	/**
	 *@author Erion Plaku
	 *@brief Add the edge connecting the indicated vertices to the
	 *       roadmap 
	 *@param idFrom index of first vertex of roadmap edge
	 *@param idTo index of second vertex of roadmap edge
	 *@param path path connecting vertex <em>idFrom</em> to
	 *       <em>idTo</em> 
	 *
	 *@par Description
	 *  The edge connecting vertex <em>idFrom</em> to vertex
	 *  <em>idTo</em> is added to the roadmap and the path
 	 *  <em>path</em> is associated with this edge. If the roadmap
	 *  is an undirected graph and the path <em>path</em> is
	 *  reversible, then the reverse edge is also added to the
	 *  roadmap. 
         */
	virtual void addEdge(const int idFrom, const int idTo, OOPSMP::Path_t path);

	/**
	 *END Connect samples and add edges to roadmap
	 *@}
	 */

    protected:
	/**
	 *@author Erion Plaku
	 *@brief Iterate over all states generated by PRM 
	 */ 
	class StateIterator : public MotionPlanner::StateIterator
	{
	public:
	    friend class PRM;

	    /**
	     *@author Erion Plaku
	     *@brief Initialize data elements to appropriate default
	     *       values 
	     */
	    StateIterator(void) : MotionPlanner::StateIterator(), m_id(0)
	    {
	    }

	    /**
	     *@author Erion Plaku
	     *@brief Free the memory allocated to different data
	     *       elements 
	     */
	    virtual ~StateIterator(void)
	    {
	    }

	protected:
	    /**
	     *@author Erion Plaku
	     *@brief Index to the current state 
	     */
	    int m_id;
	};


	/**
	 *@author Erion Plaku
	 *@brief Iterate over all paths generated by PRM 
	 */ 
	class PathIterator : public MotionPlanner::PathIterator
	{
	public:
	    friend class PRM;

	    /**
	     *@author Erion Plaku
	     *@brief Initialize data elements to appropriate default
	     *       values 
	     */
	    PathIterator(void) : MotionPlanner::PathIterator(), m_vid(0), m_eout(NULL), m_eiter(NULL)
	    {
	    }

	    /**
	     *@author Erion Plaku
	     *@brief Free the memory allocated to different data
	     *       elements 
	     */
	    virtual ~PathIterator(void)
	    {
	    }

	protected:
	    /**
	     *@author Erion Plaku
	     *@brief Index to the current roadmap vertex 
	     */
	    int m_vid;

	    /**
	     *@author Erion Plaku
	     *@brief Out edges of the current roadmap vertex
	     */
	    Map<int, GraphEdgeData_t> *m_eout;
	    
	    /**
	     *@author Erion Plaku
	     *@brief Iterator for the out edges of the current roadmap
	     *       vertex 
	     */
	    Map<int, GraphEdgeData_t>::Iterator_t  m_eiter;
	};

	/**
	 *@name Distance function for proximity computations
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Distance function used for neighbor computations 
	 *@param key1 index of the first vertex
	 *@param key2 index of the second vertex
	 *@param keyDistFnData pointer to an object of type StateSpace   
	 *
	 *@par Description
	 *  The StateSpace::distanceStates function is used for
	 *  neighbor computations 
         */
	static double proximityKeyDistFn(const OOPSMPpointer_t key1, 
					 const OOPSMPpointer_t key2, 
					 OOPSMPpointer_t       keyDistFnData)
	{
	    return ((StateSpace_const_t) keyDistFnData)->distanceStates(((VertexData_t) key1)->m_x,
									((VertexData_t) key2)->m_x);
	}

	/**
	 *END Distance function for proximity computations
	 *@}
	 */

    protected:
	/**
	 *@name Variables for preprocessing
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Variable to indicate the fraction of preprocessing
	 *       time to allocated to sample generation 
	 */
	double m_fractionSamplingTime;

	/**
	 *@author Erion Plaku
	 *@brief Minimum number of samples that should be generated --
	 *       a negative value indicates no upper bound
	 */
	int m_minNrSamples;

	/**
	 *@author Erion Plaku
	 *@brief Maximum number of samples that should be generated --
	 *       a negative value indicates no upper bound
	 */
	int m_maxNrSamples;

	/**
	 *@author Erion Plaku
	 *@brief Variable to indicate the vertex id associated with
	 *       current sample which will be connected to its
	 *       neighbors  
	 */
	int m_idConnectSample;

	/**
	 *END Variables for preprocessing
	 *@}
	 */

	/**
	 *@name Variables for roadmap and roadmap properties
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Roadmap data structure
	 */
	Graph_t m_roadmap; 

	/**
	 *@author Erion Plaku
	 *@brief Variable to indicate whether the roadmap graph is
	 *       undirected or not 
	 */
	bool m_undirected;

	/**
	 *@author Erion Plaku
	 *@brief Variable to indicate whether cycles are allowed in
	 *       the roadmap or not  
	 */
	bool m_allowCycles;

	/**
	 *@author Erion Plaku
	 *@brief Data structure for maitaining connected components of
	 *       the roadmap -- only used by an undirected roadmap
	 *       that does not allow cycles 
	 */
	DisjointSet<int> *m_disjointSet;

	/**
	 *END Variables for roadmap and roadmap properties
	 *@}
	 */

	/**
	 *@name Variables for current query solution
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Variable to indicate whether the current motion
	 *       planning query is solved or not 
	 */
	bool m_isSolved;

	/**
	 *@author Erion Plaku
	 *@brief Variable to indicate the vertex id associated with
	 *       the initial state 
	 */
	int m_idInitState;

	/**
	 *@author Erion Plaku
	 *@brief Variable to indicate the vertex id associated with
	 *       the goal state 
	 */
	int m_idGoalState;

	/**
	 *END Variables for current query solution
	 *@}
	 */

	/**
	 *@name Variables for proximity
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Data structure for computing neighbors
	 */
	ProximityKeyPointer_t m_proximity;

        /**
	 *@author Erion Plaku
	 *@brief Data structure for the current proximity query 
	 */
	ProximityQuery<OOPSMPpointer_t> m_proximityQuery;

        /**
	 *@author Erion Plaku
	 *@brief Data structure for the results of the neighbor
	 *       computations 
	 */
	ProximityResults<OOPSMPpointer_t>  m_proximityResults;

	/**
	 *@author Erion Plaku
	 *@brief Variable to indicate the number of neighbors to try
	 *       to connect the roadmap state samples 
	 */
	int  m_nrNeighborsSample;

	/**
	 *@author Erion Plaku
	 *@brief Variable to indicate the number of neighbors to try
	 *       to connect each query state to roadmap state samples 
	 */
	int m_nrNeighborsQuery;

	/**
	 *END Variables for proximity
	 *@}
	 */

	/**
	 *@name Variables for searching roadmap
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Data structure for searching roadmap graph
	 */
	SearchGraph_t m_searchGraph;

	/**
	 *END Variables for searching roadmap
	 *@}
	 */

    };

    /**
     *@author Erion Plaku
     *@brief Declare factory and specify in source file the functions
     *       that should be registered for plug-and-play functionality 
     */     
    DeclareInstanceFactory(PRM, MotionPlannerFactory);
}

#endif
