/*
 * Copyright 2008  Rice University
 * All Rights Reserved
 * 
 *       Created by Erion Plaku, Lydia Kavraki
 *       Physical and Biological Computing Group
 *       Department of Computer Science
 *       Rice University
 */

#include "MotionPlanners/PRM/OOPSMPPRM.H"
#include "Core/Path/OOPSMPPaths.H"
#include "Utils/DataStructures/SearchGraph/OOPSMPAStarSearchGraph.H"
#include "Utils/DataStructures/Map/OOPSMPFastMap.H"
#include "Utils/DataStructures/Map/OOPSMPVectorMap.H"
#include "Utils/General/OOPSMPkeywords.h"

BeginImplementInstanceFactory(PRM, MotionPlannerFactory);
RegisterFnFactory(allowCycles,             FACTORY_VAL_ARG(const bool, 0));
RegisterFnFactory(setProximity,            FACTORY_PTR_ARG(ProximityKeyPointer_t, 0));
RegisterFnFactory(setNrNeighborsSample,    FACTORY_VAL_ARG(int, 0));
RegisterFnFactory(setNrNeighborsQuery,     FACTORY_VAL_ARG(int, 0));
RegisterFnFactory(setFractionSamplingTime, FACTORY_VAL_ARG(double, 0));
RegisterFnFactory(setMinMaxNrSamples,      FACTORY_VAL_ARG(int, 0), FACTORY_VAL_ARG(int, 1));
EndImplementInstanceFactory(PRM);

/**
 *@author Erion Plaku
 *@brief Initialize data elements to appropriate default
 *       values 
 */
OOPSMP::PRM::PRM(void) : OOPSMP::MotionPlanner()
{
    m_allowCycles          = false;
    m_undirected           = true;
    m_fractionSamplingTime = 0.001;
    m_nrNeighborsSample    = 15;
    m_nrNeighborsQuery     = 15;
    m_minNrSamples         = -1;
    m_maxNrSamples         = -1;

    m_isSolved             = false;
    m_idInitState          = -1;
    m_idGoalState          = -1;	
    m_idConnectSample      = 0;	

    m_roadmap              = new Graph(new VectorMapFactory<int, GraphVertex *>(),
				       new FastMapFactory<int, GraphEdgeData *>(), NULL);
    m_disjointSet          = new DisjointSet<int>();
    m_proximity            = NULL; 
    m_searchGraph          = new AStarSearchGraph();
}

/**
 *@author Erion Plaku
 *@brief Free the memory allocated to different data elements 
 */
OOPSMP::PRM::~PRM(void)
{
    if(m_proximity)
	delete m_proximity;
    if(m_searchGraph)
	delete m_searchGraph;
    if(m_disjointSet)
	delete m_disjointSet;    
    if(m_roadmap)
	delete m_roadmap;
}

/**
 *@author Erion Plaku
 *@brief Clear the data structures and other data elements in
 *       preparation for solving motion planning queries 
 *@par Description
 * Clear the roadmap data structure, deleting the graph and
 * the vertex data and edge data associated with the vertices
 *  and edges of the graph, respectively. 
 */
void OOPSMP::PRM::clear(void)
{
    MotionPlanner::clear();
    
    if(m_proximity)
	m_proximity->clear();
    if(m_disjointSet)
	m_disjointSet->clear();
    if(m_roadmap)
	m_roadmap->clear(); 

    m_idConnectSample = 0;
    m_isSolved        = false;  
  
}

/**
 *@author Erion Plaku
 *@brief Initialize or construct data structures that will
 *       facilitate the solution of motion planning queries 
 *@param ptime maximum amount of preprocessing time allocated
 *       to motion planner 
 *@par Description
 *   Motion planner uses at most <em>ptime</em> seconds for
 *   preprocessing. PRM uses the preprocessing time to
 *   construct the roadmap.  
 */
void OOPSMP::PRM::preprocess(const double rtime)
{ 
    OOPSMPclock_t clk;
    double       t = 0.0;

    OOPSMPset_start_time(&clk);

    m_undirected = getCoreRobotData()->getPathGenerator()->isPathReversible();

    if(m_minNrSamples > 0 && m_roadmap->getNrVertices() < m_minNrSamples)
    {	
	printf("time left = %f\n", rtime);
	printf("   current and min nr. samples = %5d %5d\n", 
	       m_roadmap->getNrVertices(), m_minNrSamples);
	generateManySamples(m_minNrSamples - m_roadmap->getNrVertices(), rtime);
	printf("   current and min nr. samples = %5d %5d\n", 
	       m_roadmap->getNrVertices(), m_minNrSamples);
    }
    
    if((t = OOPSMPelapsed_time(&clk)) < rtime && m_idConnectSample < m_roadmap->getNrVertices())
    {
	printf("time left = %f\n", rtime - t);
	connectSamples(rtime - t);
    }

    if((t = OOPSMPelapsed_time(&clk)) < rtime)
    {
	printf("time left = %f\n", rtime - t);
	printf("   begin generating additional samples and connections\n");	
	generateAndConnectSamples(rtime - t);
	printf("   end   generating additional samples and connections %5d..%5d\n", 
                m_idConnectSample, m_roadmap->getNrVertices());	
    }
}

/**
 *@author Erion Plaku
 *@brief Solve the current motion planning qury
 *@param stime maximum amount of solution time allocated to
 *       motion planner 
 *
 *@par Description
 *   Motion planner attempts to solve the current motion
 *   planning query using at most  <em>stime</em> seconds.  A
 *   solution path, if found, is stored with the query. 
 */		
bool OOPSMP::PRM::solve(const double rtime)
{
    StateSpace_t ss = getCoreRobotData()->getStateSpace();

    m_idInitState = addSample(ss->copyState(getQuery()->getInitState()));
    m_idGoalState = addSample(ss->copyState(getQuery()->getGoalState()));
    
    connectSample(m_idInitState, m_nrNeighborsQuery);
    connectSample(m_idGoalState, m_nrNeighborsQuery);

    m_isSolved = m_undirected ?
	m_disjointSet->same(m_idInitState, m_idGoalState) :
	m_searchGraph->search(m_roadmap, m_idInitState, m_idGoalState);
    
    return m_isSolved;
}

/**
 *@author Erion Plaku
 *@brief Compute the solution path for the current query -- if
 *       the query is solved. 
 *
 *@remark
 *  If the current query is solved, a solution path is
 *  computed and stored as a data element of the current query
 *  using the Query::setSolutionPath function. 
 */
void OOPSMP::PRM::computeSolutionPath(void)
{
    if(m_undirected)
	m_searchGraph->search(m_roadmap, m_idInitState, m_idGoalState);

    const std::vector<int>* ipath = m_searchGraph->getPathFromStartTo(m_idGoalState);
    const int               length= ipath->size();

    if(length > 0)
    {
	Paths_t paths = new Paths(getCoreRobotData());
	for(int i = 1; i < length; i++)
	    paths->addPath(dynamic_cast<EdgeData_t>(m_roadmap->getEdgeData((*ipath)[i - 1], (*ipath)[i]))->m_path->clone());
	getQuery()->setSolutionPath(paths);
    }
}

/**
 *@author Erion Plaku
 *@brief Print states and paths of the data structure
 *       constructed to solve motion planning queries 
 *@param dt time discretization value for printing paths 
 *@param out pointer to a file open for writing
 *
 *@remark
 *  States and paths are printed using the print functions in
 *  CoreRobotData. Each path is printed as a sequence of
 *  states obtained by discretizing the path at a <em>dt</em>
 *  time interval.  
 */
void OOPSMP::PRM::printData(const double dt, FILE *out) const
{
    CoreRobotData_t    crd= getCoreRobotData();
    StateSpace_const_t ss = crd->getStateSpace();
    const int          nv = m_roadmap->getNrVertices();

    fprintf(out, "%s\n", KEYWORD_STATES);
    fprintf(out, "%d\n", nv);
    for(int i = 0; i < nv; i++)
    {
	ss->printState(dynamic_cast<VertexData_t>(m_roadmap->getVertexData(i))->m_x, out);
	fprintf(out, "\n");
    }    
    for(int i = 0; i < nv; i++)
    {
	Map<int, GraphEdgeData_t> *eout = m_roadmap->getOutEdges(i);
	for(Map<int, GraphEdgeData_t>::Iterator_t eiter = eout->beginIterator(); eiter != NULL;
	    eiter = eout->advanceIterator(eiter))
	    if(!m_undirected || i < eout->getKeyIterator(eiter))
	    {
		fprintf(out, "%s", KEYWORD_PATH);
		crd->printPath(dynamic_cast<EdgeData_t>(eout->getDataIterator(eiter))->m_path, dt, out);
	    }
    }
}

/**
 *@author Erion Plaku
 *@brief Begin iteration of paths generated by the motion
 *       planning data structure 
 *@returns
 *   Pointer to an instance of MotionPlanner::PathIterator
 *   that can be used to iterate over all the paths generated
 *   by the motion planning data structure. \n\n
 *   A NULL pointer is returned if no paths have been
 *   generated. 
 */
OOPSMP::MotionPlanner::PathIterator* OOPSMP::PRM::beginPathIterator(void) const
{
    if(m_roadmap && m_roadmap->getNrVertices() > 0)
    {
	PathIterator *pi = new PathIterator();
	return advancePathIterator(pi);
    }
    return NULL;
}

/**
 *@author Erion Plaku
 *@brief Advance the path iterator 
 *@param pi a non-NULL path iterator
 *@returns
 *  Advances the path iterator to the next path. NULL is
 *  returned if <em>pi</em> points to the last position in the
 *  path iteration and the memory associated with <em>si</em>
 *  is freed. 
 */
OOPSMP::MotionPlanner::PathIterator* OOPSMP::PRM::advancePathIterator(OOPSMP::MotionPlanner::PathIterator *pi) const
{
    PathIterator *piter = dynamic_cast<PathIterator *>(pi);
    bool          isEnd = false;
    
    if(piter->m_eiter == NULL)
    {
	piter->m_eout = m_roadmap->getOutEdges(piter->m_vid);
	piter->m_eiter= piter->m_eout->beginIterator();
    }
    else
	piter->m_eiter = piter->m_eout->advanceIterator(piter->m_eiter);
    
    while(!isEnd)
    {
	if(m_undirected == true)
	    while(piter->m_eiter != NULL && piter->m_vid > piter->m_eout->getKeyIterator(piter->m_eiter))
		piter->m_eiter = piter->m_eout->advanceIterator(piter->m_eiter);
	if(piter->m_eiter == NULL)
	{
	    piter->m_vid++;
	    if(piter->m_vid < m_roadmap->getNrVertices())
	    {
		piter->m_eout = m_roadmap->getOutEdges(piter->m_vid);
		piter->m_eiter= piter->m_eout->beginIterator();
	    }
	    else
		isEnd = true;
	}
	else
	    return piter;
	
    }
    delete piter;
    return NULL;
}

/**
 *@author Erion Plaku
 *@brief Generate and connect samples
 *@param rtime maximum amount of time 
 */
void OOPSMP::PRM::generateAndConnectSamples(const double rtime)
{
    OOPSMPclock_t clk;
    int           nrVertices  = m_roadmap->getNrVertices();
 
    OOPSMPset_start_time(&clk);
    for(double t = rtime; t > 0.0; t -= OOPSMPelapsed_time(&clk))
    {
	OOPSMPset_start_time(&clk);
	if(m_maxNrSamples < 0 || nrVertices < m_maxNrSamples)
	{
	    generateSamples(t * m_fractionSamplingTime);
	    nrVertices = m_roadmap->getNrVertices();	   
	}
	if(m_idConnectSample < nrVertices)
	    connectSamples(t - OOPSMPelapsed_time(&clk));
	
	if(m_maxNrSamples >= 0 && nrVertices >= m_maxNrSamples && m_idConnectSample >= nrVertices)
	    break;	  
    }
}

/**
 *@author Erion Plaku
 *@brief Generate valid state samples
 *@param rtime maximum amount of time used for generating
 *       state samples 
 *
 *@par Description
 *  Samples are generated using the
 *  CoreRobotData::getValidStateSampler() state sampler. Each
 *  valid state sample is added to the vertices of the
 *  roadmap. This process is repeated until <em>rtime</em> is
 *  exceeded. 
 */
void OOPSMP::PRM::generateSamples(const double rtime)
{
    CoreRobotData_t     crd = getCoreRobotData();     
    ValidStateSampler_t vss = crd->getValidStateSampler();
    StateSpace_t        ss  = crd->getStateSpace();
    State_t             x   = ss->allocState();    
    double              t   = 0.0;
    int                 nrRemaining = m_maxNrSamples >= 0 ?
	                              m_maxNrSamples - m_roadmap->getNrVertices() : -1;
    OOPSMPclock_t        clk;
    
    OOPSMPset_start_time(&clk);
    while((t = rtime - OOPSMPelapsed_time(&clk)) > 0 && nrRemaining != 0) 
  	if(vss->sample(x, t))
	{
	    addSample(x);
	    x = ss->allocState();
	    nrRemaining--;
	}
    ss->freeState(x);
}

/**
 *@author Erion Plaku
 *@brief Generate one valid state sample
 *@param rtime maximum amount of time used for generating
 *       state sample 
 *
 *@returns 
 * <em>true</em> iff a valid state sample is generated within
 * <em>rtime</em> seconds 
 *
 *@par Description
 *  The sample is generated using the
 *  CoreRobotData::getValidStateSampler() state sampler. Each
 *  valid state sample is added to the vertices of the
 *  roadmap.   
 */
bool OOPSMP::PRM::generateOneSample(const double rtime)
{
    CoreRobotData_t     crd = getCoreRobotData();     
    ValidStateSampler_t vss = crd->getValidStateSampler();
    State_t             x   = crd->getStateSpace()->allocState();    
    double              t   = 0.0;
    OOPSMPclock_t       clk;

    OOPSMPset_start_time(&clk);
    while((t = rtime - OOPSMPelapsed_time(&clk)) > 0)
  	if(vss->sample(x, t))
	{
	    addSample(x);
	    return true;
	}
    crd->getStateSpace()->freeState(x);
    return false;
}

/**
 *@author Erion Plaku
 *@brief Generate <em>n</em> valid state samples
 *@param n number of samples to generate
 *@param rtime maximum time
 */
void OOPSMP::PRM::generateManySamples(const int n, const double rtime)
{
    CoreRobotData_t     crd = getCoreRobotData();     
    ValidStateSampler_t vss = crd->getValidStateSampler();
    StateSpace_t        ss  = crd->getStateSpace();
    State_t             x   = ss->allocState();    
    int                 i   = 0;
    double              t   = 0.0;
    OOPSMPclock_t       clk;
    
    OOPSMPset_start_time(&clk);
    while(i < n && (t = OOPSMPelapsed_time(&clk)) < rtime)
  	if(vss->sample(x, rtime - t))
	{
	    addSample(x);
	    x = ss->allocState();
	    i++;
	}
    ss->freeState(x);  
}

/**
 *@author Erion Plaku
 *@brief Add state as a roadmap vertex
 *@param x state
 */
int OOPSMP::PRM::addSample(OOPSMP::State_t x)
{    
    const int   id = m_roadmap->getNrVertices();
    VertexData_t v  = new VertexData(id, x);
    
    m_roadmap->addVertex(id, v);
    m_proximity->addKey(v);
    if(m_undirected)
	m_disjointSet->make(id);     
    return id;
}

/**
 *@author Erion Plaku
 *@brief Connect neighboring roadmap state samples with valid
 *       paths 
 *@param rtime maximum amount of time used for connecting
 *       samples 
 *
 *@par Description
 *  Attempts are made to connect each roadmap state sample to 
 *  getNrNeighborsSample() neighbors, where neighbors are 
 *  computed by the Proximity data structure and connections
 *  are computed by using
 *  LocalPlanner::generateValidForwardConnectPath. \n\n
 *  This is an iterative method.  An index is maintained to
 *  indicate the current index of the roadmap vertex which
 *  will be connected to its neighbors. If after attempting
 *  these connections,  there is time left, the current index
 *  is incremented by one and this step is repeated. 
 */
void OOPSMP::PRM::connectSamples(const double rtime)
{
    printf("  begin connections %5d..%5d\n", 
	   m_idConnectSample,  m_roadmap->getNrVertices());
    
    const int    nrVertices = m_roadmap->getNrVertices();    
    OOPSMPclock_t clk;
    
    OOPSMPset_start_time(&clk);
    while(OOPSMPelapsed_time(&clk) < rtime && m_idConnectSample < nrVertices)
    {
	connectSample(m_idConnectSample, m_nrNeighborsSample);
	m_idConnectSample++;
    }

    printf("  end   connections %5d..%5d\n", m_idConnectSample, m_roadmap->getNrVertices());

}

/**
 *@author Erion Plaku
 *@brief Connect current sample to several of its neigbors
 *       with valid paths 
 *@param idFrom index of roadmap vertex
 *@param k      number of neighbors
 *
 *@par Description
 *  Attempts are made to connect the state sample associated 
 *  with the vertex with index <em>idFrom</em> to <em>k</em>
 *  neighbors, where neighbors are computed by the Proximity
 *  data structure and connections are computed by using
 *  LocalPlanner::generateValidForwardConnectPath. 
 */
void OOPSMP::PRM::connectSample(const int idFrom, const int k)
{
    if(m_proximity->isDataStructureConstructed() == false)
    {
	printf(" begin constructing proximity data structure: %d\n", m_roadmap->getNrVertices());
	m_proximity->setKeyDistFn(proximityKeyDistFn, getCoreRobotData()->getStateSpace());
	m_proximity->constructDataStructure();
	printf(" end   constructing proximity data structure\n");
    }

    LocalPlanner_t localPlanner = getCoreRobotData()->getLocalPlanner();
    VertexData_t   vFrom        = dynamic_cast<VertexData_t>(m_roadmap->getVertexData(idFrom));
    Path_t         path         = NULL;

    m_proximityQuery.setNrNeighbors(k);
    m_proximityQuery.setKey(vFrom);
    m_proximity->knearest(&m_proximityQuery, &m_proximityResults, NULL);

    const int nrResults = m_proximityResults.getNrValidResults(HUGE_VAL);
    for(int i = 0; i < nrResults; i++)
    {

	VertexData_t vTo = (VertexData_t) (m_proximityResults.getKey(i));
	if(shouldAddEdge(idFrom, vTo->m_id) &&
	   (path = localPlanner->generateValidForwardConnectPath(vFrom->m_x, vTo->m_x)))
	    addEdge(idFrom, vTo->m_id, path);
    }
}

/**
 *@author Erion Plaku
 *@brief Determine if the edge connecting the indicated
 *       vertices should be added to the roadmap 
 *@param idFrom index of first vertex of roadmap edge
 *@param idTo index of second vertex of roadmap edge
 *
 *@par Description
 *  The edge connecting vertex <em>idFrom</em> to vertex
 *  <em>idTo</em> should be added to the roadmap if cycles are
 *  allowed or if there is no path from vertex <em>idFrom</em>
 *  to <em>idTo</em> 
 */
bool OOPSMP::PRM::shouldAddEdge(const int idFrom, const int idTo)
{   
    if(m_roadmap->hasEdge(idFrom, idTo))
	return false;
    if(m_allowCycles)
	return true;
    if(m_undirected)
	return m_disjointSet->same(idFrom, idTo) == false;
    return m_searchGraph->search(m_roadmap, idFrom, idTo) == false;
}

/**
 *@author Erion Plaku
 *@brief Add the edge connecting the indicated vertices to the
 *       roadmap 
 *@param idFrom index of first vertex of roadmap edge
 *@param idTo index of second vertex of roadmap edge
 *@param path path connecting vertex <em>idFrom</em> to
 *       <em>idTo</em> 
 *
 *@par Description
 *  The edge connecting vertex <em>idFrom</em> to vertex
 *  <em>idTo</em> is added to the roadmap and the path
 *  <em>path</em> is associated with this edge. If the roadmap
 *  is an undirected graph and the path <em>path</em> is
 *  reversible, then the reverse edge is also added to the
 *  roadmap. 
 */
void OOPSMP::PRM::addEdge(const int idFrom, const int idTo, OOPSMP::Path_t path)
{   
    m_roadmap->addEdge(idFrom, idTo, new EdgeData(path));
    if(m_undirected)
    {
	assert(path->isReversible());	
	m_roadmap->addEdge(idTo, idFrom, new EdgeData(path->reverse()));
	m_disjointSet->join(idFrom, idTo);
    }
}














